from .accuracy import AccuracyWithLowRankProjectionEvaluator as AccuracyWithLowRankProjectionEvaluator
from .reconstruct import (
    ReconstructionErrorWithLowRankProjectionEvaluator as ReconstructionErrorWithLowRankProjectionEvaluator,
)



from collections.abc import Sequence

import numpy as np
import pandas as pd
import torch
from torchmetrics import MeanMetric
from tqdm.autonotebook import tqdm

from subana import intercepts
from subana.feature_map_shape_normalizers import resolve_shape_normalizer
from subana.utils.metrics import ArrayMetric

from .interface import EvaluatorWithLowRankProjection


class ReconstructionErrorWithLowRankProjectionEvaluator(EvaluatorWithLowRankProjection):
    @property
    def metric_keys(self):
        return ["norm", "recon_err", "cossim"]

    def evaluate(self, model, layer, dataloader, U, arr_ks, device="cpu", verbose=True):
        d, _ = U.shape
        n = len(arr_ks)

        metric_norm = MeanMetric()
        arr_metric_norm = ArrayMetric(
            n=n,
            base_metric=MeanMetric(),
        )
        arr_metric_recon = ArrayMetric(
            n=n,
            base_metric=MeanMetric(),
        )
        arr_metric_cosine = ArrayMetric(
            n=n,
            base_metric=MeanMetric(),
        )

        layer, feature_map_shape_normalizer = resolve_shape_normalizer(
            model=model,
            layer=layer,
            dataloader=dataloader,
            device=device,
        )

        for batch in tqdm(
            dataloader,
            desc=f"[layer={layer}] evaluating reconstruction error",
            disable=not verbose,
        ):
            x = batch[0] if isinstance(batch, Sequence) else batch

            x = x.to(device)

            output = model(x).detach().cpu()  # Ensure logits are on CPU

            norm = torch.linalg.norm(output, ord=2, dim=1)
            metric_norm.update(norm)

            for kix, k in enumerate(arr_ks):
                Uk = U[:, :k]
                hook = None
                try:
                    module = intercepts.get_module_for_layer(model=model, layer=layer)
                    hook = module.register_forward_hook(
                        intercepts.construct_fh_with_projection(
                            Uk,
                            shape_normalizer=feature_map_shape_normalizer,
                            device=device,
                        )
                    )
                    recon_output = model(x).detach().cpu()

                    np.testing.assert_equal(len(output), len(recon_output))

                    err = torch.linalg.norm(output - recon_output, ord=2, dim=1)  # Compute reconstruction error
                    arr_metric_recon.update(kix, err.cpu())

                    # fixme add cosine
                    cosine_sim = torch.nn.functional.cosine_similarity(output, recon_output, dim=1)
                    arr_metric_cosine.update(kix, cosine_sim.cpu())
                finally:
                    if hook is not None:
                        hook.remove()

        # this just make sure we have the same data types
        arr_metric_norm = np.ones(n) * float(metric_norm.compute())

        arr_metric_recon = arr_metric_recon.compute()
        arr_metric_cosine = arr_metric_cosine.compute()

        data = dict(zip(["k", *self.metric_keys], [arr_ks, arr_metric_norm, arr_metric_recon, arr_metric_cosine]))

        df = pd.DataFrame(data=data)
        df["d"] = d
        return df



import pandas as pd
from torch import nn
from torchmetrics import MeanMetric
from torchmetrics.classification import Accuracy
from tqdm.autonotebook import tqdm

from subana import intercepts
from subana.feature_map_shape_normalizers import (
    resolve_shape_normalizer,
)
from subana.utils.metrics import ArrayMetric

from .interface import EvaluatorWithLowRankProjection


class AccuracyWithLowRankProjectionEvaluator(EvaluatorWithLowRankProjection):
    def __init__(self, num_classes: int):
        self.num_classes = num_classes

    @property
    def metric_keys(self):
        return ["acc", "xent"]

    def evaluate(self, model, layer, dataloader, U, arr_ks, device="cpu", verbose=True):
        d, _ = U.shape
        n = len(arr_ks)
        arr_metric_acc = ArrayMetric(
            n=n,
            base_metric=Accuracy(num_classes=self.num_classes, task="multiclass"),
        )
        arr_metric_xent = ArrayMetric(
            n=n,
            base_metric=MeanMetric(),
        )

        layer, feature_map_shape_normalizer = resolve_shape_normalizer(
            model=model,
            layer=layer,
            dataloader=dataloader,
            device=device,
        )

        for x, y in tqdm(dataloader, desc=f"[layer={layer}] evaluating accuracy", disable=not verbose):
            x = x.to(device)

            for kix, k in enumerate(arr_ks):
                Uk = U[:, :k]

                hook = None
                try:
                    module = intercepts.get_module_for_layer(model=model, layer=layer)
                    hook = module.register_forward_hook(
                        intercepts.construct_fh_with_projection(
                            Uk,
                            shape_normalizer=feature_map_shape_normalizer,
                            device=device,
                        )
                    )

                    logits = model(x).detach().cpu()

                    xent = nn.functional.cross_entropy(logits, y, reduction="none")
                    arr_metric_acc.update(kix, logits, y)
                    arr_metric_xent.update(kix, xent)
                finally:
                    if hook is not None:
                        hook.remove()

        arr_metric_acc = arr_metric_acc.compute()
        arr_metric_xent = arr_metric_xent.compute()

        data = dict(zip(["k", *self.metric_keys], [arr_ks, arr_metric_acc, arr_metric_xent]))

        df = pd.DataFrame(data=data)
        df["d"] = d
        return df



import torch
from torch import nn

from subana import intercepts

from .interface import FeatureMapShapeNormalizer
from .mlps import MLPFeatureMapShapeNormalizer
from .vits import ViTFeatureMapShapeNormalizer


@torch.no_grad()
def resolve_shape_normalizer(
    model: nn.Module,
    layer: str,
    dataloader: torch.utils.data.DataLoader,
    device: str = "cpu",
) -> tuple[str, FeatureMapShapeNormalizer]:
    shape = intercepts.get_feature_map_shape(
        model=model,
        layer=layer,
        dataloader=dataloader,
        device=device,
    )

    if len(shape) == 4:
        return layer, None
    elif len(shape) == 2:
        return layer, MLPFeatureMapShapeNormalizer()
    elif len(shape) == 3:
        return layer, ViTFeatureMapShapeNormalizer()
    else:
        raise ValueError(f"Unsupported shape {shape} for layer {layer}. Expected 1D, 3D, or 4D tensor shapes.")  # noqa: TRY003



import typing

import torch
from torch import nn
from torch.nn import functional as F

from subana import utils
from subana.feature_map_shape_normalizers.interface import FeatureMapShapeNormalizer

ATTRIBUTE_OUTPUT_KEY = "__output"


def fh_intercept_output(mod, inp, outp: torch.Tensor):
    setattr(mod, ATTRIBUTE_OUTPUT_KEY, outp)


def fh_intercept_output_and_retain_grad(mod, inp, outp: torch.Tensor):
    outp.retain_grad()

    setattr(mod, ATTRIBUTE_OUTPUT_KEY, outp)


def get_module_output(module) -> torch.Tensor:
    return getattr(module, ATTRIBUTE_OUTPUT_KEY)


def get_module_for_layer(model: nn.Module, layer: str) -> nn.Module:
    arr_level_layers = layer.split(".")

    parent_module = model

    for attr_name in arr_level_layers:
        parsed_attr_name = utils.parse_number_if_possible(attr_name)

        if parsed_attr_name is not None:
            if not isinstance(parent_module, nn.Sequential):
                raise ValueError(f"Expected nn.Sequential, got {type(parent_module)}")  # noqa: TRY003

            if not isinstance(parsed_attr_name, int):
                raise ValueError(f"Expected integer index, got {parsed_attr_name} of type {type(parsed_attr_name)}")  # noqa: TRY003

            parent_module = parent_module[parsed_attr_name]
        else:
            parent_module = getattr(parent_module, attr_name)

    module = parent_module

    return module


def construct_fh_with_projection(
    U: torch.Tensor,
    shape_normalizer: typing.Optional[FeatureMapShapeNormalizer] = None,
    device="cpu",
) -> typing.Callable:
    d, K = U.shape

    if not (d >= K):
        raise ValueError(f"Expected d >= K, got d={d}, K={K}")  # noqa: TRY003

    UUT = (U @ U.T).unsqueeze(2).unsqueeze(3).to(device)

    def fh(mod, inp, out):
        orig_shape = out.shape

        # canonicalize the featuremap shape
        if shape_normalizer is not None:
            out = shape_normalizer.to_cnn_shape(out)

        out = F.conv2d(out, UUT)

        # decanonicalize the featuremap shape
        if shape_normalizer is not None:
            out = shape_normalizer.to_original_shape(out)
            if out.shape != orig_shape:
                raise ValueError(f"Expected shape {orig_shape}, got {out.shape}")  # noqa: TRY003

        return out.reshape(orig_shape)

    return fh


def get_feature_map_shape(
    model: nn.Module,
    layer: str,
    dataloader: torch.utils.data.DataLoader,
    device: str = "cpu",
) -> torch.Size:
    hook = None
    try:
        batch = next(iter(dataloader))
        x = utils.first_tensor_in_batch(batch)

        module = get_module_for_layer(model=model, layer=layer)
        hook = module.register_forward_hook(fh_intercept_output)

        model(x.to(device))

        output = get_module_output(module)
        shape = output.shape

    finally:
        if hook is not None:
            hook.remove()

    return shape



from copy import deepcopy

import numpy as np
import numpy.typing as npt
from torchmetrics import Metric


class ArrayMetric(Metric):
    def __init__(self, n: int, base_metric: Metric):
        self.n = n
        self.basket = []
        for _ in range(n):
            self.basket.append(deepcopy(base_metric))

    def update(self, ix: int, *args):
        self.basket[ix].update(*args)

    def compute(self) -> npt.NDArray:
        arr = []
        for i in range(self.n):
            value = self.basket[i].compute()
            arr.append(value)

        return np.array(arr, dtype=float)



import typing
from collections.abc import Sequence

import torch
from torch import nn
from torch.utils.data import DataLoader
from tqdm.autonotebook import tqdm

from subana import feature_map_shape_normalizers, intercepts, utils


def estimate_cov_mat_at_layer(
    model: nn.Module,
    layer: str,
    dataloader: DataLoader,
    device="cpu",
    pre_transform: typing.Optional[typing.Callable] = None,
) -> torch.Tensor:
    """Estimate (uncentered) covariance matrix at a given layer.

    Args:
        model (nn.Module): model to estimate covariance matrix for.
        layer (str): at which layer we want to extract the covariance matrix
        dataloader (DataLoader): use the first item in each batch as input to the model.
        device (str, optional): _description_. Defaults to "cpu".
        pre_transform (typing.Optional[typing.Callable], optional): _description_. Defaults to None.

    Raises:
        e: _description_

    Returns:
        torch.Tensor: _description_
    """

    estimator = CovarianceEstimator()

    layer, shape_normalizer = feature_map_shape_normalizers.resolve_shape_normalizer(
        model=model,
        layer=layer,
        dataloader=dataloader,
        device=device,
    )

    hook = None

    try:
        module = intercepts.get_module_for_layer(model=model, layer=layer)
        hook = module.register_forward_hook(intercepts.fh_intercept_output)

        for batch in tqdm(dataloader, desc=f"[layer={layer}] estimating covariance matrix"):
            x = batch[0] if isinstance(batch, Sequence) else batch

            x = x.to(device)
            _ = model(x)

            layer_output = intercepts.get_module_output(module)

            if pre_transform is not None:
                layer_output = pre_transform(layer_output)

            if shape_normalizer is not None:
                layer_output = shape_normalizer.to_cnn_shape(layer_output)

            estimator.update(layer_output)
    finally:
        if hook is not None:
            hook.remove()

    cov_mat = estimator.compute()
    return cov_mat


class CovarianceEstimator:
    """Batch-wisee covariance matrix estimator."""

    def __init__(self):
        self._cov_mat = None
        self._N = 0

    @torch.no_grad()
    def update(self, x: torch.Tensor):
        if len(x.shape) not in [2, 4]:
            raise ValueError(f"Expected input tensor with 2 or 4 dimensions, got shape {x.shape}")  # noqa: TRY003

        if len(x.shape) == 4:
            x = utils.flatten_4d_tensor(x)

        N, _ = x.shape

        # remark: we use cpu here to avoid memory issues when d is large.
        curr_mat = ((x.T @ x) / N).cpu()

        if self._cov_mat is None:
            self._cov_mat = curr_mat
            self._N = N
        else:
            self._cov_mat = ((self.N * self._cov_mat) + (N * curr_mat)) / (self.N + N)
            self._N = N + self.N

    @property
    def cov_mat(self) -> typing.Union[torch.Tensor, None]:
        return self._cov_mat

    @property
    def N(self) -> int:
        return self._N

    def compute(self):
        if self.cov_mat is None:
            raise ValueError("Covariance matrix has not been computed yet.")  # noqa: TRY003

        return self.cov_mat.detach()



from abc import ABC, abstractmethod

import torch


class FeatureMapShapeNormalizer(ABC):
    @abstractmethod
    def to_cnn_shape(x: torch.Tensor) -> torch.Tensor:
        pass

    @abstractmethod
    def to_original_shape(x: torch.Tensor) -> torch.Tensor:
        pass



import torch

from .interface import FeatureMapShapeNormalizer


class ViTFeatureMapShapeNormalizer(FeatureMapShapeNormalizer):
    @staticmethod
    def to_cnn_shape(x: torch.Tensor) -> torch.Tensor:
        """_summary_

        x.shape = (batch_size, num_patches, embed_dim)

        Args:
            x (torch.Tensor): _description_

        Returns:
            torch.Tensor: _description_
        """
        assert len(x.shape) == 3  # noqa: S101

        x = x.permute(0, 2, 1).unsqueeze(3)
        return x

    @staticmethod
    def to_original_shape(x: torch.Tensor) -> torch.Tensor:
        assert len(x.shape) == 4  # noqa: S101
        b, d, ntokens, w = x.shape
        assert w == 1  # noqa: S101

        x = x.squeeze(3).permute(0, 2, 1)

        return x



import re
import typing

import numpy as np
import torch
from scipy.stats import ortho_group


def get_device() -> str:
    if torch.cuda.is_available():
        return "cuda"
    else:
        return "cpu"


def first_tensor_in_batch(batch) -> torch.Tensor:
    x = batch[0] if isinstance(batch, typing.Sequence) else batch
    return x


def arange_with_grid(start: int, stop: int, step: int) -> list[int]:
    arr_ds = np.arange(start, stop, step=step)
    arr_ds_list = arr_ds.tolist()
    arr_ds_list = [*arr_ds_list, stop]

    return arr_ds_list


def eigh(cov: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
    """_summary_

    Args:
        cov (torch.Tensor): _description_

    Returns:
        typing.Tuple[torch.Tensor, torch.Tensor]: _description_
    """

    d1, d2 = cov.shape

    if d1 != d2:
        raise ValueError(f"Covariance matrix must be square, shape= {cov.shape}")  # noqa: TRY003

    eigvals, eigvecs = torch.linalg.eigh(cov)

    if len(eigvals.shape) != 1:
        raise ValueError(f"Expected 1D eigenvalues, got shape {eigvals.shape}")  # noqa: TRY003

    eigvals = torch.flip(eigvals, dims=(0,))
    eigvecs = torch.flip(eigvecs, dims=(1,))

    return eigvals, eigvecs


# fixme : use torch tensor instead
# def _solve_eigh(cov: npt.NDArray) -> typing.Tuple[NDArray, NDArray]:
#     raise NotImplementedError("use torch tensor instead")
#     """Solve eigenvalues

#     Args:
#         x (_type_): _description_

#     Returns:
#         typing.Tuple[NDArray, NDArray]: _description_
#     """

#     d1, d2 = cov.shape
#     assert d1 == d2

#     eigvals, eigvecs = np.linalg.eigh(cov)

#     assert len(eigvals.shape) == 1

#     indices = np.argsort(-eigvals)
#     eigvals: NDArray = eigvals[indices]
#     eigvecs: NDArray = eigvecs[:, indices]

#     return eigvals, eigvecs


def reshape_tensor_to_cnn_like(x: torch.Tensor) -> torch.Tensor:
    if len(x.shape) == 4:
        return x
    elif len(x.shape) == 2:
        return x.unsqueeze(2).unsqueeze(3)

    else:
        raise ValueError(f"We don't support x.shape={x.shape}")  # noqa: TRY003


def flatten_4d_tensor(x: torch.Tensor) -> torch.Tensor:
    """_summary_

    Args:
        x (torch.Tensor): _description_

    Returns:
        torch.Tensor: 2d tensor whose len(x.shape) == 2
    """

    if len(x.shape) != 4:
        raise ValueError(f"Expected 4D tensor, got shape {x.shape}")  # noqa: TRY003

    flatenned_x = x.permute(1, 0, 2, 3).flatten(start_dim=1).T

    return flatenned_x


def parse_number_if_possible(text: str) -> typing.Union[None, int]:
    is_int = re.match(r"-?\d+", text) is not None

    if is_int:
        return int(text)
    else:
        return None


def parse_layers(txt: str, arr_default_layers: list[str]) -> list[str]:
    """
    Suppose arr_default_layers=["layer1","layer2","layer3","layer4"]

    Specifying txt="@2,layer2" yields ["layer3", "layer2"].
    Here, $i is resolved to the layer given in arr_default_layers

    Args:
        txt (str): _description_
        arr_default_layers (typing.List[str]): _description_

    Returns:
        typing.List[str]: _description_
    """

    slugs = txt.split(",")

    arr_layers = []

    for slug in slugs:
        if slug[0] == "@":
            layer_ix = int(slug[1:])
            arr_layers.append(arr_default_layers[layer_ix])
        else:
            arr_layers.append(slug)

    return arr_layers


def get_random_orthogonal_matrix(d: int, seed: int) -> torch.Tensor:
    """Generate a random orthogonal matrix of size d x d.

    Args:
        d (int): Dimension of the square matrix.
        seed (int): random seed

    Returns:
        torch.Tensor: A random orthogonal matrix of shape (d, d).
    """
    rng = np.random.default_rng(seed)
    Q = ortho_group.rvs(d, random_state=rng)
    return torch.from_numpy(Q).float()



import torch

from .interface import FeatureMapShapeNormalizer


class MLPFeatureMapShapeNormalizer(FeatureMapShapeNormalizer):
    @staticmethod
    def to_cnn_shape(x: torch.Tensor) -> torch.Tensor:
        assert len(x.shape) == 2  # noqa: S101

        x = x.unsqueeze(2).unsqueeze(3)
        return x

    @staticmethod
    def to_original_shape(x: torch.Tensor) -> torch.Tensor:
        assert len(x.shape) == 4  # noqa: S101

        x = x.squeeze(3).squeeze(2)

        return x



import typing
from abc import ABC, abstractmethod

import pandas as pd
import torch
from numpy import typing as npt
from torch import nn
from torch.utils.data import DataLoader


class EvaluatorWithLowRankProjection(ABC):
    """
    Abstract base class for evaluators.
    """

    feature_map_transform: typing.Optional[typing.Callable[[list[torch.Tensor]], torch.Tensor]]

    @abstractmethod
    def evaluate(
        self,
        model: nn.Module,
        layer: str,
        dataloader: DataLoader,
        U: npt.NDArray,
        arr_ks: npt.NDArray,
        device: str = "cpu",
        verbose: bool = True,
    ) -> pd.DataFrame:
        """
        Evaluate the model or data based on the implementation.
        """
        pass

    @property
    @abstractmethod
    def metric_keys(self) -> list[str]:
        pass
